/**
 * rxrunscript
 * Run shell script in child process, Output Observable<Buffer>
 *
 * @version 5.1.1
 * @author waiting
 * @license MIT
 * @link https://github.com/waitingsong/rxrunscript#readme
 */

'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var child_process = require('child_process');
var rxjs = require('rxjs');
var operators = require('rxjs/operators');
var sharedCore = require('@waiting/shared-core');

function bindProcClose(proc) {
    const stream$ = rxjs.fromEvent(proc, 'close').pipe(
    // tap(val => console.log('bindProcClsoe data:', val)),
    operators.take(1));

    return stream$
}

function bindProcError(proc, closingNotifier$) {
    const stream$ = rxjs.fromEvent(proc, 'error').pipe(operators.take(1), 
    // tap(err => {
    //   throw err
    // }),
    operators.takeUntil(closingNotifier$));

    return stream$
}

function bindProcExit(proc) {
    const stream$ = rxjs.fromEvent(proc, 'exit').pipe(
    // tap(val => console.log('bindProcExit data:', val)),
    operators.take(1));

    return stream$
}

function bindStderrData(stderr, takeUntilNotifier$, skipUntilNofifier$, bufMaxSize) {
    const take$ = takeUntilNotifier$.pipe(operators.take(1), operators.shareReplay());
    const skip$ = skipUntilNofifier$.pipe(operators.take(1), operators.shareReplay(1));
    const event$ = bufMaxSize > 0
        ? rxjs.fromEvent(stderr, 'data')
        : rxjs.NEVER;
    const data$ = event$.pipe(operators.takeUntil(take$), operators.take(bufMaxSize), 
    // tap(buf => console.info('inner stderr1', buf.toString(), buf.byteLength)),
    operators.buffer(take$), 
    // buffer() may emit blank data, so filter
    operators.filter(arr => (arr && arr.length > 0 ? true : false)), 
    // map(Buffer.concat), // !! works not output empty array
    operators.map(arr => Buffer.concat(arr)));
    const ret$ = data$.pipe(operators.skipUntil(skip$));

    return ret$
}

function bindStdinData(stdin, inputData$) {
    const err$ = new rxjs.Subject();
    const input$ = inputData$.pipe(operators.tap(data => {
        // console.log('bindStdinData:', data)
        // debug in vsc below will cause EPIPE error
        stdin.write(data, err => {
            if (err) {
                err$.next(err);
            }
        });
    }), operators.catchError((err) => {
        err$.next(err);
        return rxjs.EMPTY
    }), operators.ignoreElements(), operators.finalize(() => {
        stdin.end();
        err$.complete();
    }));
    const ret$ = rxjs.merge(err$.asObservable(), input$).pipe(operators.tap((err) => {
        throw err
    }));

    return ret$
}

function bindStdoutData(stdout, takeUntilNotifier$) {
    const take$ = takeUntilNotifier$.pipe(operators.take(1), operators.shareReplay());
    const data$ = rxjs.fromEvent(stdout, 'data').pipe(
    // tap(buf => console.info('stdout:', buf.toString())),
    operators.takeUntil(take$));

    return data$
}

function bindEvent(proc, stderrMaxBufferSize, msgPrefixOpts, script, // for throw error
inputStream) {
    const { stderrPrefix } = msgPrefixOpts;
    const close$ = bindProcClose(proc);
    const exit$ = bindProcExit(proc);
    const closeOrExit$ = rxjs.race(close$, exit$).pipe(
    // tap(codeSignal => {
    //   console.log('close or exit event', codeSignal)
    // }),
    operators.shareReplay(1));
    const stdout$ = bindStdoutData(proc.stdout, closeOrExit$);
    const error$ = bindProcError(proc, closeOrExit$);
    const skipUntilNotifier$ = closeOrExit$.pipe(operators.mergeMap(([code]) => {
        return code === 0 || code === null ? rxjs.EMPTY : rxjs.of(void 0)
    }));
    const stderr$ = bindStderrData(proc.stderr, closeOrExit$, skipUntilNotifier$, stderrMaxBufferSize).pipe(operators.tap(buf => {
        const msg = buf.toString();

        throw new Error(`${stderrPrefix} ${script}\n${msg}`)
    }));
    const stdin$ = inputStream && typeof inputStream.subscribe === 'function'
        ? bindStdinData(proc.stdin, inputStream)
        : rxjs.EMPTY;
    const ret$ = rxjs.merge(stdout$, stderr$, stdin$, error$).pipe(operators.finalize(() => {
        // sub process may not stop
        // link: https://nodejs.org/api/child_process.html#child_process_subprocess_kill_signal
        proc.killed || proc.kill();
    }));

    return ret$
}

const initialSpawnOpts = {
    cwd: process.cwd(),
    env: Object.assign({}, process.env),
    stdio: 'pipe',
};
const initialMsgPrefixOpts = {
    errPrefix: 'Run script with error:',
    stderrPrefix: 'Run script with stderr:',
};
const initialRxRunOpts = {
    msgPrefixOpts: initialMsgPrefixOpts,
    stderrMaxBufferSize: 200,
};

function processOpts(options) {
    const { args, initialRxRunOpts: initialRxRunOpts$$1 } = options;
    const spawnOpts = processSpawnOpts(initialRxRunOpts$$1, options.spawnOpts);
    const cmd = processCommand(options.command, initialRxRunOpts$$1.msgPrefixOpts.errPrefix, spawnOpts);
    const runArgs = args && args.length ? [...args] : [];
    const errScript = runArgs.length
        ? `"${cmd} ${runArgs.join(' ')}"`
        : `"${cmd}"`;

    return {
        command: cmd,
        args: runArgs,
        spawnOpts,
        errScript,
    }
}
function processSpawnOpts(defaultRxrunOpts, options) {
    const opts = (options ? Object.assign({}, initialSpawnOpts, options) : initialSpawnOpts);

    opts.shell = true;
    /* istanbul ignore next */
    if (process.platform === 'win32') {
        opts.windowsVerbatimArguments = true;
    }
    /* istanbul ignore else */
    if (typeof opts.stderrMaxBufferSize !== 'number' || opts.stderrMaxBufferSize < 0) {
        opts.stderrMaxBufferSize = +defaultRxrunOpts.stderrMaxBufferSize;
    }
    opts.msgPrefixOpts = Object.assign({}, initialMsgPrefixOpts);
    return opts
}
function processCommand(command, errPrefix, spawnOpts) {
    let cmd = command ? command.trim() : '';

    /* istanbul ignore else */
    if (!cmd) {
        throw new TypeError(`${errPrefix}\nRun command is blank`)
    }
    const cmdLead = cmd.slice(0, 3).replace(/\\/g, '/');

    /* istanbul ignore else */
    if (spawnOpts.cwd && (cmdLead === '../' || cmdLead.slice(0, 2) === './')) {
        cmd = sharedCore.join(spawnOpts.cwd, cmd);
    }
    /* istanbul ignore else */
    if (cmd.length > 1000) {
        throw new TypeError(`${errPrefix}\nCommand length exceed 1000\n` + cmd.slice(1000) + ' ...')
    }
    return cmd
}

/**
 * Reactive Running of Command Script
 * @param command {string}
 * @param args {string[]|null}
 * @param options {Partial<RxRunOpts>|null}
 *
 * @returns Observable<Buffer>
 * @link https://nodejs.org/api/child_process.html#child_process_child_process_spawn_command_args_options
 */
function run(command, args, options) {
    const opts = processOpts({
        command, args, spawnOpts: options, initialRxRunOpts,
    });
    const { errScript } = opts;
    const { errPrefix, stderrPrefix } = opts.spawnOpts.msgPrefixOpts;
    const { inputStream } = opts.spawnOpts;
    const proc$ = runSpawn(opts.command, opts.args, opts.spawnOpts);
    const ret$ = proc$.pipe(operators.mergeMap(proc => {
        return bindEvent(proc, opts.spawnOpts.stderrMaxBufferSize, opts.spawnOpts.msgPrefixOpts, opts.errScript, inputStream)
    }), operators.catchError((err) => {
        const msg = err.message;

        if (msg.indexOf(errPrefix) !== 0 && msg.indexOf(stderrPrefix) !== 0) {
            err.message = `${errPrefix} ${errScript}\n` + msg;
        }
        throw err
    }));

    return ret$
}
function runSpawn(command, runArgs, spawnOpts) {
    try {
        const proc = child_process.spawn(command, runArgs, spawnOpts);

        return rxjs.of(proc)
    }
    catch (ex) {
        return rxjs.throwError(ex)
    }
}
function escapeShell(cmd) {
    return cmd.replace(/(["\s'$`\\])/g, '\\$1')
}

exports.run = run;
exports.escapeShell = escapeShell;
